from functools import lru_cache
from typing import Callable, Mapping

from nonebot import logger


class UnsupportedBotError(RuntimeError):
    ...


class FuncManager:
    def __init__(self, func: Mapping[str, Callable]):
        self._func = func

    def __getattr__(self, item):
        if item in self._func:
            return self._func[item]
        else:
            raise UnsupportedBotError()


class FuncManagerFactory:
    def __init__(self):
        self._registry = []

    def register(self, bot_type: str, func_name: str, func: Callable):
        self._registry.append((bot_type, func_name, func))
        logger.debug(f"registered {func_name} for {bot_type}")

    def is_supported(self, bot_type: str, func_name: str) -> bool:
        for reg in self._registry:
            if reg[0] == bot_type and reg[1] == func_name:
                return True

        return False

    @lru_cache(maxsize=8)
    def __call__(self, bot_type: str):
        func_mapping = {}
        for type_, name, func in self._registry:
            if bot_type == type_:
                func_mapping[name] = func
        return FuncManager(func_mapping)



import importlib

from .func_manager import FuncManagerFactory

platform_func = FuncManagerFactory()

supported_platform = ("onebot.v11", "kaiheila", "qqguild")

for func_name in ("is_destination_available", "handling_reaction", "get_user_nickname"):
    try:
        func_module = importlib.import_module("ssttkkl_nonebot_utils.platform." + func_name)
        for platform in supported_platform:
            try:
                func_module = importlib.import_module(f"ssttkkl_nonebot_utils.platform.{func_name}.{platform}")
                adapter_module = importlib.import_module(f"nonebot.adapters.{platform}")
                platform_func.register(adapter_module.Adapter.get_name(), func_name, getattr(func_module, func_name))
            except ImportError:
                pass
    except ImportError:
        pass

__all__ = ("platform_func",)



from nonebot import get_bot
from nonebot.adapters.onebot.v11 import Bot
from nonebot_plugin_session import Session, SessionLevel


async def get_user_nickname(session: Session) -> str:
    bot: Bot = get_bot(session.bot_id)
    if session.level == SessionLevel.LEVEL2:
        user_info = await bot.get_group_member_info(group_id=int(session.id2), user_id=int(session.id1))
        return user_info["card"] or user_info["nickname"]
    else:
        user_info = await bot.get_stranger_info(user_id=int(session.id1))
        return user_info["nickname"]



from typing import Dict, Tuple

from nonebot import logger, get_bot
from nonebot.adapters.kaiheila.exception import ActionFailed
from nonebot_plugin_session import Session, SessionLevel


async def get_guild_roles(bot_id: str, guild_id: str) -> Dict[int, int]:
    """
    获取服务器角色权限

    :param bot_id:
    :param guild_id:
    :return: 代表服务器角色权限的字典，key为role_id，value为permissions
    """
    bot = get_bot(bot_id)

    logger.debug(f"request guild {guild_id} roles")
    guild_roles = {}

    page = -1
    page_size = -1
    page_total = 2 ** 31

    while page <= page_total:
        query = {"guild_id": guild_id}
        if page != -1:
            query["page"] = page
        if page_size != -1:
            query["page_size"] = page_size

        guild_role_list = await bot.call_api("guild-role/list", query=query)
        for role in guild_role_list["items"]:
            guild_roles[role["role_id"]] = role["permissions"]

        page = guild_role_list["meta"]["page"] + 1
        page_size = guild_role_list["meta"]["page_size"]
        page_total = guild_role_list["meta"]["page_total"]

    return guild_roles


async def get_channel_overwrites(bot_id: str, channel_id: str) \
        -> Tuple[Dict[int, Tuple[bool, bool]], Dict[int, Tuple[bool, bool]]]:
    """
    获取频道覆写权限

    :param bot_id:
    :param channel_id:
    :return: {role_id: [allow], [deny]}, {user_id: [allow], [deny]}
    """
    bot = get_bot(bot_id)
    logger.debug(f"request channel {channel_id} overwrites")
    channel_roles = await bot.call_api("channel-role/index", query={"channel_id": channel_id})

    role_overwrites = {}
    for ow in channel_roles["permission_overwrites"]:
        role_id = ow["role_id"]
        role_overwrites[role_id] = ow["allow"], ow["deny"]

    user_overwrites = {}
    for ow in channel_roles["permission_users"]:
        user_id = ow["user"]["id"]
        role_overwrites[user_id] = ow["allow"], ow["deny"]

    return role_overwrites, user_overwrites


async def get_user_permissions(bot_id: str, user_id: str, channel_id: str, guild_id: str) -> int:
    """
    获取用户在该服务器该频道的权限

    :param bot_id:
    :param user_id:
    :param channel_id:
    :param guild_id:
    :return: 代表权限的掩码值（参考 https://developer.kaiheila.cn/doc/http/guild-role）
    """
    bot = get_bot(bot_id)

    logger.debug(f"request user {user_id} view in guild {guild_id}")
    user_view = await bot.call_api("user/view", query={
        "user_id": user_id,
        "guild_id": guild_id
    })

    if "roles" in user_view and len(user_view["roles"]) > 0:
        guild_roles = await get_guild_roles(bot_id, guild_id)
        role_overwrites, user_overwrites = await get_channel_overwrites(bot.self_id, channel_id)

        permissions = 0
        for role_id, role_permissions in guild_roles.items():
            # 覆写角色权限
            if role_id in role_overwrites:
                role_permissions |= role_overwrites[role_id][0]
                role_permissions &= -1 ^ role_overwrites[role_id][1]

            if role_id in user_view["roles"]:
                permissions |= role_permissions

        # 覆写用户权限
        for user_id in user_overwrites:
            if user_id == user_id:
                permissions |= user_overwrites[user_id][0]
                permissions &= -1 ^ user_overwrites[user_id][1]

        return permissions
    else:
        return 0


async def is_destination_available(session: Session) -> bool:
    try:
        if session.level != SessionLevel.LEVEL1:
            permissions = await get_user_permissions(session.bot_id,
                                                     session.id1,
                                                     session.id2,
                                                     session.id3)
            return (permissions & 4096) != 0  # 有在该频道发送消息的权限
        else:
            bot = get_bot(session.bot_id)
            # 尝试发起私聊，若能够成功发起则说明可用
            result = await bot.call_api("user-chat/create", target_id=session.id1)
            return True
    except ActionFailed:
        return False



from nonebot import get_bot
from nonebot.adapters.onebot.v11 import Bot
from nonebot.exception import ActionFailed
from nonebot_plugin_session import Session


async def is_destination_available(session: Session) -> bool:
    bot: Bot = get_bot(session.bot_id)
    try:
        if session.id2 is not None:
            group_info = await bot.get_group_info(group_id=int(session.id2))
            # 如果机器人尚未加入群, group_create_time, group_level, max_member_count 和 member_count 将会为0
            return group_info["member_count"] != 0
        else:
            friend_list = await bot.get_friend_list()
            for friend in friend_list:
                if friend["user_id"] == session.id1:
                    return True
            return False
    except ActionFailed:
        return False



from nonebot import get_bot
from nonebot.adapters.qqguild import Bot
from nonebot_plugin_session import Session


async def get_user_nickname(session: Session) -> str:
    bot: Bot = get_bot(session.bot_id)
    member = await bot.get_member(guild_id=session.id3, user_id=session.id1)
    return member.nick



from functools import wraps

from nonebot.internal.matcher import current_bot, current_event

from ..platform import platform_func


def with_handling_reaction():
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                bot = current_bot.get()
                event = current_event.get()
            except LookupError:
                await func(*args, **kwargs)
                return

            async with platform_func(bot.type).handling_reaction(bot, event):
                await func(*args, **kwargs)

        return wrapper

    return decorator



from asyncio import create_task, sleep
from contextlib import asynccontextmanager

from nonebot import logger
from nonebot.adapters.qqguild import Event, Bot, MessageEvent
from nonebot.exception import MatcherException


def add_reaction(bot: Bot, event: MessageEvent, type: int, id: str, delay: float = 0):
    async def _():
        try:
            await sleep(delay)
            await bot.put_message_reaction(channel_id=event.channel_id, message_id=event.id,
                                           type=type, id=id)
        except BaseException as e:
            logger.exception(e)

    create_task(_())


def remove_reaction(bot: Bot, event: MessageEvent, type: int, id: str, delay: float = 0):
    async def _():
        try:
            await sleep(delay)
            await bot.delete_own_message_reaction(channel_id=event.channel_id, message_id=event.id,
                                                  type=type, id=id)
        except BaseException as e:
            logger.exception(e)

    create_task(_())


@asynccontextmanager
async def handling_reaction(bot: Bot, event: Event):
    # 发送reaction时需要等待一下，否则会撞exceed frequency limit
    if not isinstance(event, MessageEvent):
        return

    add_reaction(bot, event, type=2, id="128563")  # 处理中：😳
    try:
        yield
        add_reaction(bot, event, type=2, id="128536", delay=1)  # 处理完毕：😘
    except BaseException as e:
        if not isinstance(e, MatcherException):
            add_reaction(bot, event, type=2, id="128557", delay=1)  # 处理出错：😭
        raise e
    finally:
        remove_reaction(bot, event, type=2, id="128563", delay=2)  # 处理中：😳



from nonebot import get_bot
from nonebot.adapters.kaiheila import Bot
from nonebot_plugin_session import Session


async def get_user_nickname(session: Session) -> str:
    bot: Bot = get_bot(session.bot_id)
    view = await bot.user_view(user_id=session.id1, guild_id=session.id3)
    return view.nickname



from asyncio import sleep, create_task
from contextlib import asynccontextmanager

from nonebot import logger
from nonebot.adapters.kaiheila import Event, Bot
from nonebot.adapters.kaiheila.event import PrivateMessageEvent, ChannelMessageEvent, MessageEvent
from nonebot.exception import MatcherException


def add_reaction(bot: Bot, event: Event, emoji: str, delay: float = 0):
    async def _():
        try:
            await sleep(delay)
            if isinstance(event, PrivateMessageEvent):
                await bot.directMessage_addReaction(msg_id=event.msg_id, emoji=emoji)
            elif isinstance(event, ChannelMessageEvent):
                await bot.message_addReaction(msg_id=event.msg_id, emoji=emoji)
        except BaseException as e:
            logger.exception(e)

    create_task(_())


def remove_reaction(bot: Bot, event: Event, emoji: str, delay: float = 0):
    async def _():
        try:
            await sleep(delay)
            if isinstance(event, PrivateMessageEvent):
                await bot.directMessage_deleteReaction(msg_id=event.msg_id, emoji=emoji, user_id=bot.self_id)
            elif isinstance(event, ChannelMessageEvent):
                await bot.message_deleteReaction(msg_id=event.msg_id, emoji=emoji, user_id=bot.self_id)
        except BaseException as e:
            logger.exception(e)

    create_task(_())


@asynccontextmanager
async def handling_reaction(bot: Bot, event: Event):
    if not isinstance(event, MessageEvent):
        return

    add_reaction(bot, event, ":flushed_face:")  # 处理中：😳
    try:
        yield
        add_reaction(bot, event, ":face_blowing_a_kiss:", delay=1)  # 处理完毕：😘
    except BaseException as e:
        if not isinstance(e, MatcherException):
            add_reaction(bot, event, ":loudly_crying_face:", delay=1)  # 处理出错：😭
        raise e
    finally:
        remove_reaction(bot, event, ":flushed_face:", delay=2)  # 处理中：😳



from functools import wraps

from nonebot.internal.matcher import current_matcher

from ..errors.error_handler import ErrorHandlers


def handle_error(handlers: ErrorHandlers, silently: bool = False):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            matcher = current_matcher.get()

            async def receive_error(msg: str):
                if not silently:
                    await matcher.send(msg)

            async with handlers.run_excepting(receive_error):
                return await func(*args, **kwargs)

        return wrapper

    return decorator



from contextlib import asynccontextmanager
from inspect import isawaitable
from typing import Type, Tuple, Union, Optional, Callable, List, Awaitable, Any

from nonebot import logger
from nonebot.exception import MatcherException, ActionFailed

from ssttkkl_nonebot_utils.errors.errors import BadRequestError, QueryError

T_EXCEPTABLE = Union[Type[BaseException], Tuple[Type[BaseException]]]
T_ERROR_HANDLER = Union[Callable[[BaseException], str], Callable[[BaseException], Awaitable[str]]]


class ErrorHandlers:
    def __init__(self):
        self.handlers: List[Tuple[T_EXCEPTABLE, T_ERROR_HANDLER]] = []

    def register(
            self, error_type: T_EXCEPTABLE,
            func: Optional[T_ERROR_HANDLER] = None
    ):
        def decorator(func: Optional[T_ERROR_HANDLER]):
            self.handlers.append((error_type, func))
            return func

        if func is not None:
            decorator(func)
        else:
            return decorator

    @asynccontextmanager
    async def run_excepting(self, receive_error_message: Optional[
        Union[Callable[[str], Any],
              Callable[[str], Awaitable[Any]]]
    ] = None):
        try:
            yield
        except MatcherException as e:
            raise e
        except ActionFailed as e:
            # 避免当发送消息错误时再尝试发送
            logger.exception(e)
        except (BadRequestError, QueryError) as e:
            coro = receive_error_message(e.message)
            if isawaitable(coro):
                await coro
        except BaseException as e:
            for excs, handler in self.handlers:
                if not isinstance(excs, tuple):
                    excs = (excs,)

                for exc in excs:
                    if isinstance(e, exc):
                        msg = handler(e)
                        if isawaitable(msg):
                            msg = await msg

                        if msg is not None:
                            coro = receive_error_message(msg)
                            if isawaitable(coro):
                                await coro
                        return

            logger.exception(e)
            coro = receive_error_message(f"内部错误：{type(e)}{str(e)}")
            if isawaitable(coro):
                await coro



import asyncio
from asyncio import sleep, shield, create_task
from contextlib import asynccontextmanager

from nonebot import logger
from nonebot.adapters.onebot.v11 import Event, Bot


async def send_delayed_loading_prompt(bot: Bot, event: Event):
    try:
        await sleep(3)

        logger.debug(f"send delayed loading")
        await shield(bot.send(event, "努力加载中"))
    except asyncio.CancelledError as e:
        raise e
    except BaseException as e:
        logger.exception(e)


@asynccontextmanager
async def handling_reaction(bot: Bot, event: Event):
    task = create_task(send_delayed_loading_prompt(bot, event))

    try:
        yield
    finally:
        if task and not task.done():
            task.cancel()
